@@ -1,19 +1,7 @@
 /*
- *  Copyright (c) 2007 Boudewijn Rempt <boud@valdyas.org>
+ *  SPDX-FileCopyrightText: 2007 Boudewijn Rempt <boud@valdyas.org>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_base_node.h"
@@ -22,10 +10,16 @@
 #include <kis_image.h>
 #include <kis_icon.h>
 #include <KoProperties.h>
+#include <KisAnimatedOpacityProperty.h>
 #include <KoColorSpace.h>
 #include <KoCompositeOpRegistry.h>
+
+#include <QSharedPointer>
+#include "kis_pointer_utils.h"
+
 #include "kis_paint_device.h"
 #include "kis_layer_properties_icons.h"
+#include "kis_default_bounds_node_wrapper.h"
 
 #include "kis_scalar_keyframe_channel.h"
 
@@ -36,34 +30,29 @@ struct Q_DECL_HIDDEN KisBaseNode::Private
     KisBaseNode::Property hack_visible; //HACK
     QUuid id;
     QMap<QString, KisKeyframeChannel*> keyframeChannels;
-    QScopedPointer<KisScalarKeyframeChannel> opacityChannel;
+    KisAnimatedOpacityProperty opacityProperty;
 
-    bool systemLocked;
-    bool collapsed;
-    bool supportsLodMoves;
-    bool animated;
-    bool useInTimeline;
+    bool collapsed {false};
+    bool supportsLodMoves {false};
+    bool animated {false};
+    bool pinnedToTimeline {false};
     KisImageWSP image;
 
-    Private(KisImageWSP image)
+    Private(KisImageWSP p_image)
         : id(QUuid::createUuid())
-        , systemLocked(false)
-        , collapsed(false)
-        , supportsLodMoves(false)
-        , animated(false)
-        , useInTimeline(false)
-        , image(image)
+        , opacityProperty(new KisDefaultBounds(p_image), &properties, OPACITY_OPAQUE_U8)
+        , image(p_image)
     {
     }
 
     Private(const Private &rhs)
         : compositeOp(rhs.compositeOp),
           id(QUuid::createUuid()),
-          systemLocked(false),
+          opacityProperty(new KisDefaultBounds(rhs.image), &properties, OPACITY_OPAQUE_U8),
           collapsed(rhs.collapsed),
           supportsLodMoves(rhs.supportsLodMoves),
           animated(rhs.animated),
-          useInTimeline(rhs.useInTimeline),
+          pinnedToTimeline(rhs.pinnedToTimeline),
           image(rhs.image)
     {
         QMapIterator<QString, QVariant> iter = rhs.properties.propertyIterator();
@@ -92,6 +81,8 @@ KisBaseNode::KisBaseNode(KisImageWSP image)
     setSupportsLodMoves(true);
 
     m_d->compositeOp = COMPOSITE_OVER;
+
+    connect(&m_d->opacityProperty, SIGNAL(changed(quint8)), this, SIGNAL(opacityChanged(quint8)));
 }
 
 
@@ -100,10 +91,12 @@ KisBaseNode::KisBaseNode(const KisBaseNode & rhs)
     , KisShared()
     , m_d(new Private(*rhs.m_d))
 {
-    if (rhs.m_d->opacityChannel) {
-        m_d->opacityChannel.reset(new KisScalarKeyframeChannel(*rhs.m_d->opacityChannel, 0));
-        m_d->keyframeChannels.insert(m_d->opacityChannel->id(), m_d->opacityChannel.data());
+    if (rhs.m_d->opacityProperty.hasChannel()) {
+        m_d->opacityProperty.transferKeyframeData(rhs.m_d->opacityProperty);
+        m_d->keyframeChannels.insert(m_d->opacityProperty.channel()->id(), m_d->opacityProperty.channel());
     }
+
+    connect(&m_d->opacityProperty, SIGNAL(changed(quint8)), this, SIGNAL(opacityChanged(quint8)));
 }
 
 KisBaseNode::~KisBaseNode()
@@ -111,39 +104,20 @@ KisBaseNode::~KisBaseNode()
     delete m_d;
 }
 
-KisPaintDeviceSP KisBaseNode::colorPickSourceDevice() const
+KisPaintDeviceSP KisBaseNode::colorSampleSourceDevice() const
 {
     return projection();
 }
 
 quint8 KisBaseNode::opacity() const
 {
-    if (m_d->opacityChannel) {
-        qreal value = m_d->opacityChannel->currentValue();
-
-        if (!qIsNaN(value)) {
-            return value;
-        }
-    }
-
-    return nodeProperties().intProperty("opacity", OPACITY_OPAQUE_U8);
+    return m_d->opacityProperty.get();
 }
 
 void KisBaseNode::setOpacity(quint8 val)
 {
-    if (m_d->opacityChannel) {
-        KisKeyframeSP activeKeyframe = m_d->opacityChannel->currentlyActiveKeyframe();
-
-        if (activeKeyframe) {
-            m_d->opacityChannel->setScalarValue(activeKeyframe, val);
-        }
-    }
-
-    if (opacity() == val) return;
-
-    setNodeProperty("opacity", val);
-
-    baseNodeInvalidateAllFramesCallback();
+    m_d->opacityProperty.set(val);
+    baseNodeChangedCallback();
 }
 
 quint8 KisBaseNode::percentOpacity() const
@@ -221,8 +195,10 @@ bool KisBaseNode::check(const KoProperties & properties) const
 }
 
 
-QImage KisBaseNode::createThumbnail(qint32 w, qint32 h)
+QImage KisBaseNode::createThumbnail(qint32 w, qint32 h, Qt::AspectRatioMode aspectRatioMode)
 {
+    Q_UNUSED(aspectRatioMode);
+
     try {
         QImage image(w, h, QImage::Format_ARGB32);
         image.fill(0);
@@ -233,9 +209,15 @@ QImage KisBaseNode::createThumbnail(qint32 w, qint32 h)
 
 }
 
-QImage KisBaseNode::createThumbnailForFrame(qint32 w, qint32 h, int time)
+int KisBaseNode::thumbnailSeqNo() const
 {
-    Q_UNUSED(time)
+    return -1;
+}
+
+QImage KisBaseNode::createThumbnailForFrame(qint32 w, qint32 h, int time, Qt::AspectRatioMode aspectRatioMode)
+{
+    Q_UNUSED(time);
+    Q_UNUSED(aspectRatioMode);
     return createThumbnail(w, h);
 }
 
@@ -267,6 +249,36 @@ bool KisBaseNode::userLocked() const
     return m_d->properties.boolProperty(KisLayerPropertiesIcons::locked.id(), false);
 }
 
+bool KisBaseNode::belongsToIsolatedGroup() const
+{
+    if (!m_d->image) {
+        return false;
+    }
+
+    const KisBaseNode* element = this;
+
+    while (element) {
+        if (element->isIsolatedRoot()) {
+            return true;
+        } else {
+            element = element->parentCallback().data();
+        }
+    }
+
+    return false;
+}
+
+bool KisBaseNode::isIsolatedRoot() const
+{
+    if (!m_d->image) {
+        return false;
+    }
+
+    const KisBaseNode* isolatedRoot = m_d->image->isolationRootNode().data();
+
+    return (this == isolatedRoot);
+}
+
 void KisBaseNode::setUserLocked(bool locked)
 {
     const bool isLocked = m_d->properties.boolProperty(KisLayerPropertiesIcons::locked.id(), true);
@@ -280,7 +292,7 @@ bool KisBaseNode::isEditable(bool checkVisibility) const
 {
     bool editable = true;
     if (checkVisibility) {
-        editable = (visible(false) && !userLocked());
+        editable = ((visible(false) || belongsToIsolatedGroup()) && !userLocked());
     }
     else {
         editable = (!userLocked());
@@ -302,7 +314,13 @@ bool KisBaseNode::hasEditablePaintDevice() const
 
 void KisBaseNode::setCollapsed(bool collapsed)
 {
+    const bool oldCollapsed = m_d->collapsed;
+
     m_d->collapsed = collapsed;
+
+    if (oldCollapsed != collapsed) {
+        baseNodeCollapsedChangedCallback();
+    }
 }
 
 bool KisBaseNode::collapsed() const
@@ -341,9 +359,15 @@ bool KisBaseNode::supportsLodMoves() const
     return m_d->supportsLodMoves;
 }
 
+bool KisBaseNode::supportsLodPainting() const
+{
+    return true;
+}
+
 void KisBaseNode::setImage(KisImageWSP image)
 {
     m_d->image = image;
+    m_d->opacityProperty.updateDefaultBounds(new KisDefaultBounds(image));
 }
 
 KisImageWSP KisBaseNode::image() const
@@ -376,6 +400,19 @@ KisKeyframeChannel * KisBaseNode::getKeyframeChannel(const QString &id) const
     return i.value();
 }
 
+bool KisBaseNode::isPinnedToTimeline() const
+{
+    return m_d->pinnedToTimeline;
+}
+
+void KisBaseNode::setPinnedToTimeline(bool pinned)
+{
+   if (pinned == m_d->pinnedToTimeline) return;
+
+   m_d->pinnedToTimeline = pinned;
+   baseNodeChangedCallback();
+}
+
 KisKeyframeChannel * KisBaseNode::getKeyframeChannel(const QString &id, bool create)
 {
     KisKeyframeChannel *channel = getKeyframeChannel(id);
@@ -402,45 +439,47 @@ void KisBaseNode::enableAnimation()
     baseNodeChangedCallback();
 }
 
-bool KisBaseNode::useInTimeline() const
-{
-    return m_d->useInTimeline;
-}
-
-void KisBaseNode::setUseInTimeline(bool value)
-{
-    if (value == m_d->useInTimeline) return;
-
-    m_d->useInTimeline = value;
-    baseNodeChangedCallback();
-}
-
 void KisBaseNode::addKeyframeChannel(KisKeyframeChannel *channel)
 {
     m_d->keyframeChannels.insert(channel->id(), channel);
-    emit keyframeChannelAdded(channel);
+    Q_EMIT keyframeChannelAdded(channel);
 }
 
 KisKeyframeChannel *KisBaseNode::requestKeyframeChannel(const QString &id)
 {
     if (id == KisKeyframeChannel::Opacity.id()) {
-        Q_ASSERT(m_d->opacityChannel.isNull());
+        Q_ASSERT(!m_d->opacityProperty.hasChannel());
 
         KisPaintDeviceSP device = original();
+        KisNode* node = dynamic_cast<KisNode*>(this);
 
-        if (device) {
-            KisScalarKeyframeChannel * channel = new KisScalarKeyframeChannel(
-                KisKeyframeChannel::Opacity,
-                0, 255,
-                device->defaultBounds(),
-                KisKeyframe::Linear
-            );
-
-            m_d->opacityChannel.reset(channel);
-
-            return channel;
+        if (device && node) {
+            m_d->opacityProperty.makeAnimated(node);
+            return m_d->opacityProperty.channel();
         }
     }
 
     return 0;
 }
+
+bool KisBaseNode::supportsKeyframeChannel(const QString &id)
+{
+    if (id == KisKeyframeChannel::Opacity.id() && original()) {
+        return true;
+    }
+
+    return false;
+}
+
+QDebug operator<<(QDebug dbg, const KisBaseNode::Property &prop)
+{
+    dbg.nospace() << "Property(" << prop.id << ", " << prop.state;
+
+    if (prop.isInStasis) {
+        dbg.nospace() << ", in-stasis";
+    }
+
+    dbg.nospace() << ")";
+
+    return dbg.space();
+}
