@@ -1,24 +1,12 @@
 /*
- *  Copyright (c) 2015 Dmitry Kazakov <dimula73@gmail.com>
+ *  SPDX-FileCopyrightText: 2015 Dmitry Kazakov <dimula73@gmail.com>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "kis_regenerate_frame_stroke_strategy.h"
 
-#include <QRegion>
+#include <KisRegion.h>
 #include "kis_image_interfaces.h"
 #include "kis_image_animation_interface.h"
 #include "kis_node.h"
@@ -28,6 +16,7 @@
 #include "kis_full_refresh_walker.h"
 #include "kis_async_merger.h"
 #include "kis_projection_updates_filter.h"
+#include <KisLockFrameGenerationLock.h>
 
 
 struct KisRegenerateFrameStrokeStrategy::Private
@@ -35,9 +24,10 @@ struct KisRegenerateFrameStrokeStrategy::Private
     Type type;
     int frameId;
     int previousFrameId;
-    QRegion dirtyRegion;
+    KisRegion dirtyRegion;
     KisImageAnimationInterface *interface;
-    KisProjectionUpdatesFilterSP prevUpdatesFilter;
+    QStack<KisProjectionUpdatesFilterSP> prevUpdatesFilters;
+    std::optional<KisLockFrameGenerationLock> frameGenerationLock;
 
     class Data : public KisStrokeJobData {
     public:
@@ -61,8 +51,10 @@ struct KisRegenerateFrameStrokeStrategy::Private
         if (!image) {
             return;
         }
-        prevUpdatesFilter = image->projectionUpdatesFilter();
-        image->setProjectionUpdatesFilter(KisProjectionUpdatesFilterSP());
+
+        while (KisProjectionUpdatesFilterCookie cookie = image->currentProjectionUpdatesFilter()) {
+            prevUpdatesFilters.push(image->removeProjectionUpdatesFilter(cookie));
+        }
     }
 
     void restoreUpdatesFilter() {
@@ -70,15 +62,19 @@ struct KisRegenerateFrameStrokeStrategy::Private
         if (!image) {
             return;
         }
-        image->setProjectionUpdatesFilter(prevUpdatesFilter);
-        prevUpdatesFilter.clear();
+
+        while (!prevUpdatesFilters.isEmpty()) {
+            image->addProjectionUpdatesFilter(prevUpdatesFilters.pop());
+        }
     }
 };
 
 KisRegenerateFrameStrokeStrategy::KisRegenerateFrameStrokeStrategy(int frameId,
-                                                                   const QRegion &dirtyRegion,
-                                                                   KisImageAnimationInterface *interface)
-    : KisSimpleStrokeStrategy("regenerate_external_frame_stroke"),
+                                                                   const KisRegion &dirtyRegion,
+                                                                   bool isCancellable,
+                                                                   KisImageAnimationInterface *interface,
+                                                                   KisLockFrameGenerationLock &&frameGenerationLock)
+    : KisSimpleStrokeStrategy(QLatin1String("regenerate_external_frame_stroke")),
       m_d(new Private)
 {
     m_d->type = EXTERNAL_FRAME;
@@ -86,6 +82,7 @@ KisRegenerateFrameStrokeStrategy::KisRegenerateFrameStrokeStrategy(int frameId,
     m_d->frameId = frameId;
     m_d->dirtyRegion = dirtyRegion;
     m_d->interface = interface;
+    m_d->frameGenerationLock = std::move(frameGenerationLock);
 
     enableJob(JOB_INIT);
     enableJob(JOB_FINISH, true, KisStrokeJobData::BARRIER);
@@ -98,17 +95,17 @@ KisRegenerateFrameStrokeStrategy::KisRegenerateFrameStrokeStrategy(int frameId,
 
     setRequestsOtherStrokesToEnd(false);
     setClearsRedoOnStart(false);
-    setCanForgetAboutMe(true);
+    setCanForgetAboutMe(isCancellable);
 }
 
 KisRegenerateFrameStrokeStrategy::KisRegenerateFrameStrokeStrategy(KisImageAnimationInterface *interface)
-    : KisSimpleStrokeStrategy("regenerate_current_frame_stroke", kundo2_i18n("Render Animation")),
+    : KisSimpleStrokeStrategy(QLatin1String("regenerate_current_frame_stroke"), kundo2_i18n("Render Animation")),
       m_d(new Private)
 {
     m_d->type = CURRENT_FRAME;
 
     m_d->frameId = 0;
-    m_d->dirtyRegion = QRegion();
+    m_d->dirtyRegion = KisRegion();
     m_d->interface = interface;
 
     enableJob(JOB_INIT);
@@ -135,12 +132,18 @@ void KisRegenerateFrameStrokeStrategy::initStrokeCallback()
         return;
     }
     if (m_d->type == EXTERNAL_FRAME) {
+        if (!image->animationInterface()->hasAnimation()) {
+            warnKrita << "KisRegenerateFrameStrokeStrategy::initStrokeCallback(): WARNING: trying to"
+                      << "regenerate an external frame on a non-animated image, that will cause"
+                      << "a useless consumption of memory";
+        }
+
         m_d->saveAndResetUpdatesFilter();
         image->disableUIUpdates();
         m_d->interface->saveAndResetCurrentTime(m_d->frameId, &m_d->previousFrameId);
     } else if (m_d->type == CURRENT_FRAME) {
         m_d->interface->blockFrameInvalidation(true);
-        m_d->interface->updatesFacade()->refreshGraphAsync(KisNodeSP());
+        m_d->interface->updatesFacade()->refreshGraphAsync(KisNodeSP(), KisProjectionUpdateFlag::DontInvalidateFrames);
     }
 }
 
@@ -151,7 +154,9 @@ void KisRegenerateFrameStrokeStrategy::doStrokeCallback(KisStrokeJobData *data)
     KIS_ASSERT(!m_d->dirtyRegion.isEmpty());
     KIS_ASSERT(m_d->type == EXTERNAL_FRAME);
 
-    KisBaseRectsWalkerSP walker = new KisFullRefreshWalker(d->cropRect);
+    const bool skipNonRenderableNodes = m_d->type == EXTERNAL_FRAME;
+    KisBaseRectsWalkerSP walker = new KisFullRefreshWalker(d->cropRect,
+                                                           skipNonRenderableNodes ? KisFullRefreshWalker::SkipNonRenderableNodes : KisFullRefreshWalker::None);
     walker->collectRects(d->root, d->rect);
 
     KisAsyncMerger merger;
@@ -164,12 +169,15 @@ void KisRegenerateFrameStrokeStrategy::finishStrokeCallback()
     if (!image) {
         return;
     }
+
+
     if (m_d->type == EXTERNAL_FRAME) {
         m_d->interface->notifyFrameReady();
         m_d->interface->restoreCurrentTime(&m_d->previousFrameId);
         image->enableUIUpdates();
         m_d->restoreUpdatesFilter();
     } else if (m_d->type == CURRENT_FRAME) {
+        m_d->interface->notifyFrameRegenerated();
         m_d->interface->blockFrameInvalidation(false);
     }
 }
@@ -200,7 +208,7 @@ KisStrokeStrategy* KisRegenerateFrameStrokeStrategy::createLodClone(int levelOfD
      */
     return m_d->type == CURRENT_FRAME ?
         new KisRegenerateFrameStrokeStrategy(m_d->interface) :
-        new KisSimpleStrokeStrategy("dumb-lodn-KisRegenerateFrameStrokeStrategy");
+        new KisSimpleStrokeStrategy(QLatin1String("dumb-lodn-KisRegenerateFrameStrokeStrategy"));
 }
 
 void KisRegenerateFrameStrokeStrategy::suspendStrokeCallback()
