@@ -1,21 +1,9 @@
 /*
  *  tool_transform_args.h - part of Krita
  *
- *  Copyright (c) 2010 Marc Pegon <pe.marc@free.fr>
+ *  SPDX-FileCopyrightText: 2010 Marc Pegon <pe.marc@free.fr>
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *  SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "tool_transform_args.h"
@@ -28,33 +16,19 @@
 
 #include "kis_liquify_transform_worker.h"
 #include "kis_dom_utils.h"
+#include <QMatrix4x4>
 
 
 ToolTransformArgs::ToolTransformArgs()
-    : m_mode(FREE_TRANSFORM)
-    , m_defaultPoints(true)
-    , m_origPoints {QVector<QPointF>()}
-    , m_transfPoints {QVector<QPointF>()}
-    , m_warpType(KisWarpTransformWorker::RIGID_TRANSFORM)
-    , m_alpha(1.0)
-    , m_transformedCenter(QPointF(0, 0))
-    , m_originalCenter(QPointF(0, 0))
-    , m_rotationCenterOffset(QPointF(0, 0))
-    , m_aX(0)
-    , m_aY(0)
-    , m_aZ(0)
-    , m_scaleX(1.0)
-    , m_scaleY(1.0)
-    , m_shearX(0.0)
-    , m_shearY(0.0)
-    , m_liquifyProperties(new KisLiquifyProperties())
-    , m_pixelPrecision(8)
-    , m_previewPixelPrecision(16)
+    : m_liquifyProperties(new KisLiquifyProperties())
 {
     KConfigGroup configGroup =  KSharedConfig::openConfig()->group("KisToolTransform");
     QString savedFilterId = configGroup.readEntry("filterId", "Bicubic");
     setFilterId(savedFilterId);
     m_transformAroundRotationCenter = configGroup.readEntry("transformAroundRotationCenter", "0").toInt();
+    m_meshShowHandles = configGroup.readEntry("meshShowHandles", true);
+    m_meshSymmetricalHandles = configGroup.readEntry("meshSymmetricalHandles", true);
+    m_meshScaleHandles = configGroup.readEntry("meshScaleHandles", false);
 }
 
 void ToolTransformArgs::setFilterId(const QString &id) {
@@ -100,18 +74,38 @@ void ToolTransformArgs::init(const ToolTransformArgs& args)
     m_editTransformPoints = args.m_editTransformPoints;
     m_pixelPrecision = args.pixelPrecision();
     m_previewPixelPrecision = args.previewPixelPrecision();
+    m_externalSource = args.externalSource();
 
     if (args.m_liquifyWorker) {
         m_liquifyWorker.reset(new KisLiquifyTransformWorker(*args.m_liquifyWorker.data()));
     }
 
+    m_meshTransform = args.m_meshTransform;
+    m_meshShowHandles = args.m_meshShowHandles;
+    m_meshSymmetricalHandles = args.m_meshSymmetricalHandles;
+    m_meshScaleHandles = args.m_meshScaleHandles;
+
     m_continuedTransformation.reset(args.m_continuedTransformation ? new ToolTransformArgs(*args.m_continuedTransformation) : 0);
 }
 
+bool ToolTransformArgs::meshScaleHandles() const
+{
+    return m_meshScaleHandles;
+}
+
+void ToolTransformArgs::setMeshScaleHandles(bool meshScaleHandles)
+{
+    m_meshScaleHandles = meshScaleHandles;
+
+    KConfigGroup configGroup =  KSharedConfig::openConfig()->group("KisToolTransform");
+    configGroup.writeEntry("meshScaleHandles", meshScaleHandles);
+}
+
 void ToolTransformArgs::clear()
 {
     m_origPoints.clear();
     m_transfPoints.clear();
+    m_meshTransform = KisBezierTransformMesh();
 }
 
 ToolTransformArgs::ToolTransformArgs(const ToolTransformArgs& args)
@@ -163,9 +157,12 @@ bool ToolTransformArgs::operator==(const ToolTransformArgs& other) const
         m_editTransformPoints == other.m_editTransformPoints &&
         (m_liquifyProperties == other.m_liquifyProperties ||
          *m_liquifyProperties == *other.m_liquifyProperties) &&
+        m_meshTransform == other.m_meshTransform &&
 
         // pointer types
 
+        m_externalSource == other.m_externalSource &&
+
         ((m_filter && other.m_filter &&
           m_filter->id() == other.m_filter->id())
          || m_filter == other.m_filter) &&
@@ -217,6 +214,8 @@ bool ToolTransformArgs::isSameMode(const ToolTransformArgs& other) const
              *m_liquifyWorker == *other.m_liquifyWorker)
             || m_liquifyWorker == other.m_liquifyWorker;
 
+    } else if (m_mode == MESH) {
+        result &= m_meshTransform == other.m_meshTransform;
     } else {
         KIS_SAFE_ASSERT_RECOVER_NOOP(0 && "unknown transform mode");
     }
@@ -236,7 +235,8 @@ ToolTransformArgs::ToolTransformArgs(TransformMode mode,
                                      double alpha,
                                      bool defaultPoints,
                                      const QString &filterId,
-                                     int pixelPrecision, int previewPixelPrecision)
+                                     int pixelPrecision, int previewPixelPrecision,
+                                     KisPaintDeviceSP externalSource)
     : m_mode(mode)
     , m_defaultPoints(defaultPoints)
     , m_origPoints {QVector<QPointF>()}
@@ -257,6 +257,7 @@ ToolTransformArgs::ToolTransformArgs(TransformMode mode,
     , m_liquifyProperties(new KisLiquifyProperties())
     , m_pixelPrecision(pixelPrecision)
     , m_previewPixelPrecision(previewPixelPrecision)
+    , m_externalSource(externalSource)
 {
     setFilterId(filterId);
 }
@@ -267,23 +268,56 @@ ToolTransformArgs::~ToolTransformArgs()
     clear();
 }
 
-void ToolTransformArgs::translate(const QPointF &offset)
+void ToolTransformArgs::translateSrcAndDst(const QPointF &offset)
 {
-    if (m_mode == FREE_TRANSFORM || m_mode == PERSPECTIVE_4POINT) {
-        m_originalCenter += offset;
-        m_rotationCenterOffset += offset;
-        m_transformedCenter += offset;
+    transformSrcAndDst(QTransform::fromTranslate(offset.x(), offset.y()));
+}
+
+void ToolTransformArgs::transformSrcAndDst(const QTransform &t)
+{
+    if (m_mode == FREE_TRANSFORM ) {
+        m_originalCenter = t.map(m_originalCenter);
+        m_transformedCenter = t.map(m_transformedCenter);
+
+        QMatrix4x4 m(t);
+        m_cameraPos = m * m_cameraPos;
+    } else if (m_mode == PERSPECTIVE_4POINT) {
+        m_originalCenter = t.map(m_originalCenter);
+        m_transformedCenter = t.map(m_transformedCenter);
+
+        m_flattenedPerspectiveTransform = t.inverted() * m_flattenedPerspectiveTransform * t;
+
     } else if(m_mode == WARP || m_mode == CAGE) {
         for (auto &pt : m_origPoints) {
-            pt += offset;
+            pt = t.map(pt);
         }
 
+        for (auto &pt : m_transfPoints) {
+            pt = t.map(pt);
+        }
+    } else if (m_mode == LIQUIFY) {
+        KIS_ASSERT_RECOVER_RETURN(m_liquifyWorker);
+        m_liquifyWorker->transformSrcAndDst(t);
+    } else if (m_mode == MESH) {
+        m_meshTransform.transformSrcAndDst(t);
+    } else {
+        KIS_ASSERT_RECOVER_NOOP(0 && "unknown transform mode");
+    }
+}
+
+void ToolTransformArgs::translateDstSpace(const QPointF &offset)
+{
+    if (m_mode == FREE_TRANSFORM || m_mode == PERSPECTIVE_4POINT) {
+        m_transformedCenter += offset;
+    } else if(m_mode == WARP || m_mode == CAGE) {
         for (auto &pt : m_transfPoints) {
             pt += offset;
         }
     } else if (m_mode == LIQUIFY) {
         KIS_ASSERT_RECOVER_RETURN(m_liquifyWorker);
-        m_liquifyWorker->translate(offset);
+        m_liquifyWorker->translateDstSpace(offset);
+    } else if (m_mode == MESH) {
+        m_meshTransform.translate(offset);
     } else {
         KIS_ASSERT_RECOVER_NOOP(0 && "unknown transform mode");
     }
@@ -306,14 +340,20 @@ bool ToolTransformArgs::isIdentity() const
 
         return true;
     } else if (m_mode == LIQUIFY) {
-        // Not implemented!
-        return false;
+        return !m_liquifyWorker || m_liquifyWorker->isIdentity();
+    } else if (m_mode == MESH) {
+        return m_meshTransform.isIdentity();
     } else {
         KIS_ASSERT_RECOVER_NOOP(0 && "unknown transform mode");
         return true;
     }
 }
 
+bool ToolTransformArgs::isUnchanging() const
+{
+    return !m_externalSource && isIdentity();
+}
+
 void ToolTransformArgs::initLiquifyTransformMode(const QRect &srcRect)
 {
     m_liquifyWorker.reset(new KisLiquifyTransformWorker(srcRect, 0, 8));
@@ -380,6 +420,11 @@ void ToolTransformArgs::toXML(QDomElement *e) const
 
         m_liquifyProperties->toXML(&liqEl);
         m_liquifyWorker->toXML(&liqEl);
+    } else if (m_mode == MESH) {
+        QDomElement meshEl = doc.createElement("mesh_transform");
+        e->appendChild(meshEl);
+
+        KisDomUtils::saveValue(&meshEl, "mesh", m_meshTransform);
     } else {
         KIS_ASSERT_RECOVER_RETURN(0 && "Unknown transform mode");
     }
@@ -460,7 +505,7 @@ ToolTransformArgs ToolTransformArgs::fromXML(const QDomElement &e)
         if(args.m_mode == CAGE){
             // Pixel precision is a parameter introduced in Krita 4.2, so we should
             // expect it not being present in older files. In case it is not found,
-            // just use the defalt value initialized by c-tor (that is, do nothing).
+            // just use the default value initialized by c-tor (that is, do nothing).
 
             (void) KisDomUtils::loadValue(warpEl, "pixelPrecision", &args.m_pixelPrecision);
             (void) KisDomUtils::loadValue(warpEl, "previewPixelPrecision", &args.m_previewPixelPrecision);
@@ -480,6 +525,14 @@ ToolTransformArgs ToolTransformArgs::fromXML(const QDomElement &e)
 
         *args.m_liquifyProperties = KisLiquifyProperties::fromXML(e);
         args.m_liquifyWorker.reset(KisLiquifyTransformWorker::fromXML(e));
+    } else if (args.m_mode == MESH) {
+        QDomElement meshEl;
+
+        result =
+            KisDomUtils::findOnlyElement(e, "mesh_transform", &meshEl);
+
+        result &= KisDomUtils::loadValue(meshEl, "mesh", &args.m_meshTransform);
+
     } else {
         KIS_ASSERT_RECOVER_NOOP(0 && "Unknown transform mode");
     }
@@ -510,3 +563,57 @@ const ToolTransformArgs* ToolTransformArgs::continuedTransform() const
 {
     return m_continuedTransformation.data();
 }
+
+const KisBezierTransformMesh *ToolTransformArgs::meshTransform() const
+{
+    return &m_meshTransform;
+}
+
+KisBezierTransformMesh *ToolTransformArgs::meshTransform()
+{
+    return &m_meshTransform;
+}
+
+bool ToolTransformArgs::meshShowHandles() const
+{
+    return m_meshShowHandles;
+}
+
+void ToolTransformArgs::setMeshShowHandles(bool value)
+{
+    m_meshShowHandles = value;
+
+    KConfigGroup configGroup =  KSharedConfig::openConfig()->group("KisToolTransform");
+    configGroup.writeEntry("meshShowHandles", value);
+}
+
+bool ToolTransformArgs::meshSymmetricalHandles() const
+{
+    return m_meshSymmetricalHandles;
+}
+
+void ToolTransformArgs::setMeshSymmetricalHandles(bool value)
+{
+    m_meshSymmetricalHandles = value;
+
+    KConfigGroup configGroup =  KSharedConfig::openConfig()->group("KisToolTransform");
+    configGroup.writeEntry("meshSymmetricalHandles", value);
+}
+
+void ToolTransformArgs::scale3dSrcAndDst(qreal scale)
+{
+    const QTransform t = QTransform::fromScale(scale, scale);
+
+    if (m_mode == FREE_TRANSFORM ) {
+        m_originalCenter = t.map(m_originalCenter);
+        m_transformedCenter = t.map(m_transformedCenter);
+
+        // we need to scale Z-coordinate of the camera pos as well,
+        // so we cannot just do `QMatrix4x4 m(t)`.
+        QMatrix4x4 m;
+        m.scale(scale);
+        m_cameraPos = m * m_cameraPos;
+    } else {
+        transformSrcAndDst(t);
+    }
+}
